@@ -16,10 +16,12 @@
 #   limitations under the License.
 #
 
+__version__ = "1.0.11"
+
 
 def _main():
     from kubernator import app
-    return app.main()
+    return app.main() or 0
 
 
 def main():
